import crypto from 'node:crypto';

/*
 * Copyright (c) 2020-2023 Filipe Guerra
 * https://github.com/openwar/mongodb-queue
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */

// some helper functions
function id() {
  return crypto.randomBytes(16).toString('hex');
}
class MongoDbQueueImpl {
  get collection() {
    return this._db.collection(this._name);
  }
  constructor(db, name, options = {}) {
    if (!db) {
      throw new Error('Please provide a mongodb.MongoClient.db');
    }
    if (!name) {
      throw new Error('Please provide a queue name');
    }
    this._db = db;
    this._name = name;
    this._visibility = options.visibility || 30;
  }
  async createIndexes() {
    await this.collection.createIndex({
      deleted: 1,
      visible: 1
    });
    await this.collection.createIndex({
      ack: 1
    }, {
      unique: true,
      sparse: true
    });
  }
  async add(payload, options) {
    const hashKey = options?.hashKey;
    const delay = options?.delay ?? 0;
    const now = Date.now();
    const insertFields = {
      createdAt: new Date(now),
      visible: new Date(now + delay * 1000),
      payload,
      tries: 0
    };
    if (hashKey === undefined) {
      const result = await this.collection.insertOne({
        ...insertFields,
        occurrences: 1
      });
      return result.insertedId.toHexString();
    }
    let filter = {
      payload: {
        $eq: hashKey
      }
    };
    if (typeof payload === 'object') {
      filter = {
        [`payload.${String(hashKey)}`]: payload[hashKey]
      };
    }
    const message = await this.collection.findOneAndUpdate(filter, {
      $inc: {
        occurrences: 1
      },
      $set: {
        updatedAt: new Date()
      },
      $setOnInsert: insertFields
    }, {
      upsert: true,
      returnDocument: 'after',
      includeResultMetadata: true
    });
    if (!message.value) {
      throw new Error(`Queue.add(): Failed add message`);
    }
    return message.value._id.toHexString();
  }
  async get(options = {}) {
    const visibility = options.visibility || this._visibility;
    const now = Date.now();
    const query = {
      deleted: null,
      visible: {
        $lte: new Date(now)
      }
    };
    const update = {
      $inc: {
        tries: 1
      },
      $set: {
        updatedAt: new Date(now),
        ack: id(),
        visible: new Date(now + visibility * 1000)
      }
    };
    const result = await this.collection.findOneAndUpdate(query, update, {
      sort: {
        _id: 1
      },
      returnDocument: 'after',
      includeResultMetadata: true
    });
    const message = result.value;

    // nothing in the queue
    if (!message) return;
    if (!message.ack || !message.updatedAt) {
      throw new Error(`Queue.get(): Failed to update message`);
    }

    // convert to an external representation
    return {
      id: message._id.toHexString(),
      ack: message.ack,
      createdAt: message.createdAt,
      updatedAt: message.updatedAt,
      payload: message.payload,
      tries: message.tries,
      occurrences: message.occurrences ?? 1
    };
  }
  async ping(ack, options = {}) {
    const now = Date.now();
    const visibility = options.visibility || this._visibility;
    const query = {
      ack: ack,
      visible: {
        $gt: new Date(now)
      },
      deleted: null
    };
    const update = {
      $set: {
        visible: new Date(now + visibility * 1000)
      }
    };
    const message = await this.collection.findOneAndUpdate(query, update, {
      returnDocument: 'after',
      includeResultMetadata: true
    });
    if (!message.value) {
      throw new Error(`Queue.ping(): Unidentified ack : ${ack}`);
    }
    return message.value._id.toHexString();
  }
  async ack(ack) {
    const now = Date.now();
    const query = {
      ack: ack,
      visible: {
        $gt: new Date(now)
      },
      deleted: null
    };
    const update = {
      $set: {
        deleted: new Date(now)
      }
    };
    const message = await this.collection.findOneAndUpdate(query, update, {
      returnDocument: 'after',
      includeResultMetadata: true
    });
    if (!message.value) {
      throw new Error(`Queue.ack(): Unidentified ack : ${ack}`);
    }
    return message.value._id.toHexString();
  }
  async total() {
    return await this.collection.countDocuments();
  }
  async size() {
    const query = {
      deleted: null,
      visible: {
        $lte: new Date()
      }
    };
    return await this.collection.countDocuments(query);
  }
  async inFlight() {
    const query = {
      ack: {
        $exists: true
      },
      visible: {
        $gt: new Date()
      },
      deleted: null
    };
    return await this.collection.countDocuments(query);
  }
  async done() {
    const query = {
      deleted: {
        $exists: true
      }
    };
    return await this.collection.countDocuments(query);
  }
}
function mongoDbQueue(db, name, options = {}) {
  return new MongoDbQueueImpl(db, name, options);
}

export { mongoDbQueue as default };
//# sourceMappingURL=mongodb-queue.es.js.map
